// ALab.io v1.0: java package for ALab to import and export AIFF and WAVE soundfiles
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab.io;

import ALab.*;
import java.io.*;

/** This class is used to import and export AIFF format.
 * <br><br><IMG SRC="../point.gif">
 * @author  Michael Bliem
 * @version 1.0 */
public class AIFF { private AIFF() {}

  /** Writes all values of <code>in</code> to soundfile <code>out</code>. The
   * container is copied before used to allow multiple <code>write()</code>.
   * @param in   the container with the stored <code>AObject</code>s.
   * @param rate the samplerate.
   * @param bit  the bitrate (1..8bit, 2..16bit, 3..24bit, 4..32bit).
   * @param out  the soundfile to be written.
   * @exception AException if container is empty, IOError occurs or
   * any calculation error occurs. */
  static public void write(AContainer in, int rate, int bit, OutputStream out) throws AException {
    try {
      in = (AContainer)in.copy();
      DataOutputStream o = new DataOutputStream(out);
      long samples = in.getLength();
      int  channel = in.getChannel();
      long data    = samples * bit * channel;
      o.writeBytes("FORM");           // Group ID
      o.writeInt((int)(data + 46));   // File Size
      o.writeBytes("AIFF");           // Type ID

      o.writeBytes("COMM");           // Chunk ID
      o.writeInt(18);                 // Chunk Size
      o.writeShort((short)channel);   // channels
      o.writeInt((int)samples);       // samples
      o.writeShort((short)(bit * 8)); // bits
      o.write(extBytes((double)rate));// samplerate

      o.writeBytes("SSND");           // Chunk ID
      o.writeInt((int)(data + 8));    // ChunkSize
      o.writeInt(0);                  // offset
      o.writeInt(0);                  // blocksize
      in.write(rate, bit, true, o);   // data
      o.close(); }
    catch(IOException e) { throw new AException (
      "AIFF.write(): could not write AIFF file."); }}

  /** Returns an <code>ASample</code> instance using the values of input
   * <code>in</code>. All parameters are set depending on the header of the
   * input soundfile <code>in</code>. <code>start</code> is set to zero,
   * <code>length</code> is set to the amount of samples in the soundfile.
   * The input stream is copied before used.
   * @param in the soundfile.
   * @exception AException if type is not valid or IOError occurs. */
  static public ASample getInput(AInputStream in) throws AException {
    return getWave(in).getInput(); }

  /** Returns an <code>ASampleLoop</code> instance using the values of input
   * <code>in</code>. All parameters are set depending on the header of the
   * input soundfile <code>in</code>. <code>start</code> is set to zero,
   * <code>length</code> is set to the amount of samples in the soundfile.
   * <code>freq</code> is set to samplerate divided by <code>length</code>.
   * The input stream is copied before used.
   * @param in the soundfile.
   * @exception AException if type is not valid or IOError occurs. */
  static public ASampleLoop getWave(AInputStream in) throws AException {
    try {
      int  channel = 2;
      int  bit = 2;
      int  rate = 44100;
      long samples = 0;
      in = in.copy();
      DataInputStream i = new DataInputStream((InputStream)in);
      if(!readID(i).equals("FORM")) throw new AException(
        "AIFF.getWave(): group ID is not FORM.");
      i.skip(4);
      if(!readID(i).equals("AIFF")) throw new AException(
        "AIFF.getWave(): type ID is not AIFF.");
      long offset = 12;
      boolean comm = false;
      boolean ssnd = false;
      int length;
      while(!comm || !ssnd) {
        String s = readID(i);
        length = i.readInt();
        offset += 8;
        if(s.equals("COMM")) {
          if(comm) throw new AException(
            "AIFF.getWave(): second COMM chunk found.");
          comm = true;
          channel = (int)(i.readShort());
          samples = (long)(i.readInt());
          bit = (i.readShort() + 7) / 8;
          byte[] b = new byte[10];
          i.readFully(b);
          rate = (int)extDouble(b);
          i.skip(length - 18);
          offset += length; }
        else if(s.equals("SSND")) {
          if(ssnd) throw new AException(
            "AIFF.getWave(): second SSND chunk found.");
          ssnd = true;
          offset += (i.readInt() + 8);
          long l = length - 4;
          while(l > 0) l -= i.skip(l); }
        else { i.skip(length); offset += length; }}
      in.setPos(offset);
      ASample s = new ASample(channel, in, bit, true, 0, samples);
      return new ASampleLoop(s, (double)rate / samples, 0, 0, samples);
    } catch(IOException e) { throw new AException(
      "AIFF.getWave(): error reading AIFF format."); }}

  static private String readID(DataInputStream in) throws IOException {
    byte[] b = new byte[4]; in.readFully(b); return new String(b); }

  static private byte[] extBytes(double v) {
    byte[] b = new byte[10];
    long bits = Double.doubleToLongBits(v);
    long mantisse = ((bits & 0x000fffffffffffffL)
                           | 0x0010000000000000L) <<  11;
    long exponent = ((bits & 0x8000000000000000L) >>> 48)
                  |(((bits & 0x7ff0000000000000L) >>  52) + 15360);
    b[0] = (byte)((exponent & 0xff00L) >>> 8);
    b[1] = (byte)(exponent & 0xffL);
    for(int i = 2; i < 10; i++)
      b[i] = (byte)((mantisse >> ((9 - i) * 8)) & 0xff);
    return b; }

  static private double extDouble(byte[] b) {
    long exponent = ((b[0] & 0x80L) << 56)
          |((((long)((short)(((b[0] & 0xff) << 8) | b[1]))) - 15360) << 52);
    long mantisse = b[2] & 0xffL;
    for(int i = 3; i < 10; i++) mantisse = (mantisse << 8) | (b[i] & 0xffL);
    mantisse = (mantisse >> 11) & 0x000fffffffffffffL;
    return Double.longBitsToDouble(exponent | mantisse); }
}