// ALab.io v1.0: java package for ALab to import and export AIFF and WAVE soundfiles
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab.io;

import ALab.*;
import java.io.*;

/** This class is used to import and export uncompressed WAVE format.
 * <br><br><IMG SRC="../point.gif">
 * @author  Michael Bliem
 * @version 1.0 */
public class WAVE { private WAVE() {}

  /** Writes all values of <code>in</code> to soundfile <code>out</code>. The
   * container is copied before used to allow multiple <code>write()</code>.
   * @param in   the container with the stored <code>AObject</code>s.
   * @param rate the samplerate.
   * @param bit  the bitrate (1..8bit, 2..16bit, 3..24bit, 4..32bit).
   * @param out  the soundfile to be written.
   * @exception AException if container is empty, IOError occurs or
   * any calculation error occurs. */
  static public void write(AContainer in, int rate, int bit, OutputStream out) throws AException {
    try {
      in = (AContainer)in.copy();
      DataOutputStream o = new DataOutputStream(out);
      long samples = in.getLength();
      int  channel = in.getChannel();
      int  size    = bit * channel;
      long data    = samples * bit * channel;
      o.writeBytes("RIFF");                 // Group ID
      o.write(intBytes((int)(data + 36)));  // File Size
      o.writeBytes("WAVE");                 // Type ID

      o.writeBytes("fmt ");                 // Chunk ID
      o.write(intBytes(16));                // Chunk Size
      o.write(shortBytes(1));               // no compression
      o.write(shortBytes((short)channel));  // channels
      o.write(intBytes(rate));              // samplerate
      o.write(intBytes(rate * size));       // bytes per second
      o.write(shortBytes((short)(size)));   // samplesize
      o.write(shortBytes((short)(bit * 8)));// bits

      o.writeBytes("data");                 // Chunk ID
      o.write(intBytes((int)(data)));       // ChunkSize
      in.write(rate, bit, false, o);        // data
      o.close(); }
    catch(IOException e) { throw new AException (
      "WAVE.write(): could not write WAVE file."); }}

  static private byte[] intBytes(int v) {
    ByteArrayOutputStream b = new ByteArrayOutputStream();
    DataOutputStream f = new DataOutputStream(b);
    try { f.writeInt(v); } catch(IOException e) {}
    byte[] res = new byte[4];
    byte[] dat = b.toByteArray();
    for(int i = 0; i < 4; i++) res[3-i] = dat[i];
    return res; }

  static private byte[] shortBytes(int v) {
    ByteArrayOutputStream b = new ByteArrayOutputStream();
    DataOutputStream f = new DataOutputStream(b);
    try { f.writeShort(v); } catch(IOException e) {}
    byte[] res = new byte[2];
    byte[] dat = b.toByteArray();
    for(int i = 0; i < 2; i++) res[1-i] = dat[i];
    return res; }

  /** Returns an <code>ASample</code> instance using the values of input
   * <code>in</code>. All parameters are set depending on the header of the
   * input soundfile <code>in</code>. <code>start</code> is set to zero,
   * <code>length</code> is set to the amount of samples in the soundfile.
   * The input stream is copied before used.
   * @param in the soundfile.
   * @exception AException if type is not valid or IOError occurs. */
  static public ASample getInput(AInputStream in) throws AException {
    return getWave(in).getInput(); }

  /** Returns an <code>ASampleLoop</code> instance using the values of input
   * <code>in</code>. All parameters are set depending on the header of the
   * input soundfile <code>in</code>. <code>start</code> is set to zero,
   * <code>length</code> is set to the amount of samples in the soundfile.
   * <code>freq</code> is set to samplerate divided by <code>length</code>.
   * The input stream is copied before used.
   * @param in the soundfile.
   * @exception AException if type is not valid or IOError occurs. */
  static public ASampleLoop getWave(AInputStream in) throws AException {
    try {
      int  channel = 2;
      int  bit = 2;
      int  rate = 44100;
      long samples = 0;
      in = in.copy();
      DataInputStream i = new DataInputStream((InputStream)in);
      if(!readID(i).equals("RIFF")) throw new AException(
        "WAVE.getWave(): group ID is not RIFF.");
      i.skip(4);
      if(!readID(i).equals("WAVE")) throw new AException(
        "WAVE.getWave(): type ID is not WAVE.");
      long offset = 12;
      boolean fmt_ = false;
      boolean data = false;
      int length;
      while(!fmt_ || !data) {
        String s = readID(i);
        length = readInt(i);
        offset += 8;
        if(s.equals("fmt ")) {
          if(fmt_) throw new AException(
            "WAVE.getWave(): second fmt chunk found.");
          fmt_ = true;
          if(readShort(i) != 1) throw new AException(
            "WAVE.getWave(): unable to handle compressed WAVE file.");
          channel = (int)(readShort(i));
          rate = readInt(i);
          i.skip(6);
          bit = (readShort(i) + 7) / 8;
          i.skip(length - 16);
          offset += length; }
        else if(s.equals("data")) {
          if(data) throw new AException(
            "WAVE.getWave(): second data chunk found.");
          data = true;
          samples = length;
          long l = length;
          while(l > 0) l -= i.skip(l); }
        else { i.skip(length); offset += length; }}
      in.setPos(offset);
      samples /= channel * bit;
      ASample s = new ASample(channel, in, bit, false, 0, samples);
      return new ASampleLoop(s, (double)rate / samples, 0, 0, samples);
    } catch(IOException e) { throw new AException(
      "WAVE.getWave(): error reading WAVE format."); }}

  static private String readID(DataInputStream in) throws IOException {
    byte[] b = new byte[4]; in.readFully(b); return new String(b); }

  static private int readInt(DataInputStream in) throws IOException {
    byte[] b = new byte[4]; in.readFully(b);
    int ret = b[3] & 0xff;
    for(int i = 2; i >= 0; i--) ret = (ret << 8) | (b[i] & 0xff);
    return ret; }

  static private short readShort(DataInputStream in) throws IOException {
    byte[] b = new byte[2]; in.readFully(b);
    return (short)(((b[1] & 0xff) << 8) | (b[0] & 0xff)); }
}