// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

import java.io.*;
import java.util.*;

/**
 * This class is a container for several {@link ALab.AObject AObject}s.
 * AObjects can be appended with a certain length to an AContainer. These
 * AObjects are played one after another. The first appended AObject will
 * be played at first, the last appended will be played at last. If there
 * is no more AObject in the container, the output is zero. AObjects are
 * not stored as they are, but as {@link ALab.APart APart}s, because AParts
 * make AObjects to have a cartain length. If any appended AObject has less
 * channels than this container, zero is played instead of the missing
 * channel. If any AObject has more channels, these additional channels are
 * ignored.<br><br>
 * AContainer can write its AObjects to an output stream. The samples are
 * decoded in PCM-format and one value per channel (eg: a stereo AObject with
 * left values  <code>0 0.5 1</code> and right values <code>0 -0.5 -1</code>.
 * These values are transformed to 16Bit values: L: <code>0 16384 32768</code>
 * and R: <code>0 -16384 -32768</code>. The last L-value is above the maximal
 * positive value and therefor decremented to <code>32767</code>. Then these
 * values are transformed to binary-code (big endian format (motorola)
 * notated in HEX-code here): L: <code>0000 4000 7FFF</code> and R:
 * <code>0000 C000 8000</code>. Now one value of the left channel is printed
 * to output followed by one of the second. This results in the following
 * output: <code>0000 0000 4000 C000 7FFF 8000</code>).
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see     ALab.APart APart
 */
public class AContainer extends AObject {
  private Vector  container = new Vector();
  private APart   cPart     = null;
  private int     cIndex    = 0;
  private long    length    = 0;
  private boolean end       = false;
  private double  ampFactor = 0;
  private long    time      = 0;
  private long    vps       = 0;

  /** Creates a new container.
   * @param ch the amount of audio-channels. */
  public AContainer(int ch) { super(ch); }

  /** Creates a new <code>AContainer</code>
   * by copying the given <code>AContainer</code>.
   * @param o the given object. */
  protected AContainer(AContainer o) { super(o);
    setContainer(o.getContainer()); }

  /** Returns all AObjects stored in this container.
   * @return an {@link java.util.Enumeration Enumeration}
   *         representing all stored AObjects. */
  public Enumeration getContainer() { return container.elements(); }

  /** Fills container with all AObjects stored in the given
   * {@link java.util.Enumeration Enumeration}.
   * @param e an {@link java.util.Enumeration Enumeration}
   *         representing all stored AObjects. */
  protected void setContainer(Enumeration e) {
    container = new Vector(); while(e.hasMoreElements()) {
      APart p = (APart)e.nextElement();
      append(p.getObj(), p.getLength()); }}

  /** Appends one AObject with a certain length to the container.
   * The AObject <code>o</code> is stored as an APart with
   * the length <code>length</code>.
   * @param o the AObject to be stored.
   * @param length the amount of samples that should be stored
                   (one is not possible and interpreted as two).
   * @see       ALab.APart APart */
  public void append(AObject o, long length) {
    APart p = new APart(o, length);
    container.addElement(p); this.length+= p.getLength(); }

  /** Returns length of this container. */
  public long getLength() { return length; }

  protected double getValue(int ch, int rate) throws AException {
    if(!end && isNewSample(ch) && ((cPart == null) || cPart.endReached())) {
      if(cIndex >= container.size()) end = true;
      else while((cPart = (APart)(container.elementAt(cIndex++))).getLength() <= 0); }
    if(end) return 0; else return cPart.nextValue(ch, rate); }

  /** Writes all AObjects to an output stream. Additional information
   * is written to standard output (eg: samples written per second).
   * This method only works correctly the first time it is invoced.
   * Every trial to invoce it more than once will result in silence.
   * @param rate the sample rate (eg: 44100).
   * @param bit bitrange in byte (eg: 2 for 16Bit), maximal 4bytes.
   * @param format <code>true:</code> big endian (motorola).
   *               <code>false:</code> little endian (intel).
   * @param out the output stream.
   * @exception AException if container is empty, IOError occurs or
   * any calculation error occurs. */
  public void write(int rate, int bit, boolean format, OutputStream out) throws AException {
    if(length <= 0) throw new AException(
      "AContainer.write(): nothing in container to write.");
    if(bit < 1) bit = 1; else if(bit > 4) bit = 4;
    int bitMax = (1 << (8*bit-1)) - 1;
    int valueMax = 0;
    for(long i = 0; i < length; i++) { status(i);
      for(int ch = 0; ch < channel; ch++) {
        int value = (int)(nextValue(ch, rate) * bitMax);
        valueMax = Math.max(valueMax, Math.abs(value));
        try { out.write(intToByteArray(value, bit, format), 0, bit); }
        catch(IOException e) { throw new AException(
          "AContainer.write(): error writing to OutputStream."); }
    }} ampFactor = (double)bitMax/valueMax; status(length); }

  private void status(long i) {
    if(i == 0) {
      System.out.print("samples per second: 0.     ");
      time = System.currentTimeMillis(); }
    else if(i == length) {
      while((vps/=10) > 0) System.out.print("\b");
      vps = (i * 1000) / Math.max(1, System.currentTimeMillis() - time);
      System.out.println("\b\b\b\b\b\b\b"+vps+".     ");
      System.out.println("you can increase amp by: "+ampFactor+"."); }
    else if((i % 10000) == 0) {
      while((vps/=10) > 0) System.out.print("\b");
      vps = (i * 1000) / Math.max(1, System.currentTimeMillis() - time);
      System.out.print("\b\b\b\b\b\b\b"+vps+".     "); }}

  private static byte[] intToByteArray(int value, int size, boolean m) {
    byte[] b = new byte[size];
    if(!m && (size == 1)) { b[0] = (byte)(value + 128); return b; }
    for(int i = 0; i < size; i++)
      if(m) b[size - (1 + i)] = (byte)((value >> (8 * i)) & 0xff);
      else b[i] = (byte)((value >> (8 * i)) & 0xff);
    return b; }

  public AObject copy() { return new AContainer(this); }
}
