// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

/**
 * A class dividing the values of two {@link ALab.AObject}s.
 * If the division is out of range (<code>-1..1</code>)
 * zero is returned. Due to high values resulting if a
 * value is divided by a value near zero, a limit can be set
 * to tell what is the highest value that should be in
 * range. eg:
 * <pre>
 * limit = 10.0;
 * value1 = 0.5;           // inside range -1..1
 * value2 = 0.1;           // inside range -1..1
 * div = value1 / value2;  // result: 5 is out of range
 * div = div / limit;      // result: 0.5 is in range
 * </pre>
 * This limit value can be modulated. If a division by zero
 * occurs, zero is returned.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 */
public class ADiv extends AObject {
  private double  limit;
  private AObject limitMod;
  private AObject obj1;
  private AObject obj2;

  /** Creates a new object-adding instance.
   * @param obj1 the first object to be divided.
   * @param obj2 the second object to be divided.
   * @param limit the limit. */
  public ADiv(AObject obj1, AObject obj2, double limit) {
    super(Math.max(obj1.getChannel(),obj2.getChannel()));
    setObj1(obj1); setObj2(obj2); setLimit(limit); }

  /** Creates a new <code>ADiv</code>
   * by copying the given <code>ADiv</code>.
   * @param o the given object. */
  protected ADiv(ADiv o) {  super(o);
    setObj1(o.getObj1());   setObj2(o.getObj2());
    setLimit(o.getLimit()); setLimitMod(o.getLimitMod()); }

  /** Returns the first object that is divided. */
  public AObject getObj1() { return obj1; }

  /** Sets the first object that is divided
   * to a copy of <code>o</code>.
   * @param o the object. */
  protected void setObj1(AObject o) { obj1 = o.copy(); }

  /** Returns the second object that is divided. */
  public AObject getObj2() { return obj2; }

  /** Sets the second object that is divided
   * to a copy of <code>o</code>.
   * @param o the object. */
  protected void setObj2(AObject o) { obj2 = o.copy(); }

  /** Returns the limit. */
  public double getLimit() { return limit; }

  /** Sets the limit.
   * @param v the limit value. */
  public void setLimit(double v) { limit = v; }

  /** Returns the limit modulator. */
  public AObject getLimitMod() { return limitMod; }

  /** Sets the limit modulator to a copy of <code>o</code>.
   * @param o the limit modulator. */
  public void setLimitMod(AObject o) {
    if(o != null) limitMod = o.copy(); }

  protected double getValue(int ch, int rate) throws AException {
    double v;
    double l = nextObjValue(limitMod, ch, rate, limit);
    double f = obj1.nextValue(ch, rate);
    double s = obj2.nextValue(ch, rate);
    if((s == 0) || (l == 0)) return 0; else v = f / (s * l);
    if(Math.abs(v) > 1) v = 0; return v; }

  public AObject copy() { return new ADiv(this); }
}
