// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

/**
 * A logarithmic waveform generator. If the value of
 * <code>width</code> is low, the waveform is relatively flat
 * (linear). If it is high, the waveform is very sharp (negative
 * values are interpreted as positive values).<br><br>
 * <IMG SRC="space.gif" WIDTH=50 HEIGHT=1>
 * <IMG SRC="ALog.gif" ALT="Diagram of the width effect."><br>
 * <br>This value can be modulated by the <code>widthMod</code>.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 */
public class ALog extends AWave {
  private double  factor = 2.0 / Math.PI;
  private double  width;
  private AObject widthMod = null;

  /**
   * Creates a logarithmic waveform generator.
   * @param ch    the amount of audio channels.
   * @param freq  the frequency.
   * @param phase the phase.
   * @param width the flattness or sharpness of the waveform.
   * @see ALab.AWave AWave
   */
  public ALog(int ch, double freq, double phase, double width) {
    super(ch, freq, phase); setWidth(width); }

  /** Creates a new <code>ALog</code>
   * by copying the given <code>ALog</code>.
   * @param w the given object. */
  protected ALog(ALog w) { super(w);
    setWidth(w.getWidth()); setWidthMod(w.getWidthMod()); }

  /** Returns the width of the waveform. */
  public double getWidth() { return width; }

  /** Sets the width of the waveform. If <code>v</code> is a
   * value near zero, the wave form is very flat, if the value
   * is higher, the waveform becomes sharp.
   * @param v the flattnes or sharpness of the waveform. */
  public void setWidth(double v) { width = v; }

  /** Returns the width modulator. */
  public AObject getWidthMod() { return widthMod; }

  /** Sets the width modulator to a copy of <code>o</code>.
   * @param o the width modulator. */
  public void setWidthMod(AObject o) {
    if(o != null) widthMod = o.copy(); }

  protected double getValue(int ch, int rate) throws AException {
    double w = Math.abs(nextObjValue(widthMod, ch, rate, width));
    if(w < 0.000000001) w = 0.000000001;
    double p = nextPhase(ch, rate);
    double f = Math.exp(-w);
    double m = p * factor;
    if     (m >=  1) m = 2 - m;
    else if(m >=  0) ;
    else if(m >= -1) m*= -1;
    else             m+=  2;
    double v = Math.log((Math.exp(w) - f) * m + f) / (w*2) + 0.5;
    if(p < 0) return -v; else return v; }

  public AObject copy() { return new ALog(this); }
}