// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

/**
 * Abstract basisclass for most other classes in ALab. Subclasses have to
 * implement the methods <code>{@link #getValue(int, int) getValue()}</code>
 * and <code>{@link #copy() copy()}</code>. Every AObject has a certain
 * amount of audio channels. These channels are numbered, starting at zero
 * (eg. an AObject with two audio-channels: the method <code>{@link
 * #getChannel() getChannel()}</code> returns <code>2</code>. These two
 * channels are numberes with channel zero for the first and channel one for
 * the second channel). Every AObject must have at least one channel.
 * <br><br>
 * In this documentation a 'sample' always means all values that mean to be
 * listened at the same time, but at a different channel (eg. AObject with
 * two channels: a sample consists of two values).
 * <br><br>
 * To write the values of an AObject to an OutputStream this object must be
 * appended to an {@link ALab.AContainer AContainer} and the method {@link
 * ALab.AContainer#write(int, int, boolean, OutputStream) write()} must be
 * called. <br><br>
 * AObjects only throw {@link ALab.AException AExceptions}.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see ALab.AContainer AContainer
 * @see ALab.AException AException
 */
public abstract class AObject {

  /** The amount of audio channels. */
  protected int channel;

  /** The last channel number. The last channel called by {@link
   * #nextValue(int, int) nextValue()}, or {@link #channel channel}
   * if <code>nextValue()</code> never has been called. */
  protected int last;

  /** Creates a new object with a certain amount of channels. Does nothing
   * else but calling the method {@link #setChannel(int) setChannel()}.
   * @param ch the amount of channels. */
  protected AObject(int ch) { setChannel(ch); }

  /** Creates a new object with as much channels as the given object.
   * Simply calls the method {@link #setChannel(int) setChannel()}.<br><br>
   * Subclasses of AObject always should implement a constructor like this
   * one, building itself by using the fields of the given object and using
   * copys of the objects the given object uses. This constructor can
   * be used for implementing the method {@link #copy() copy()}.<br><br>
   * @param o the given object. */
  protected AObject(AObject o) { setChannel(o.getChannel()); }

  /** Returns the amount of {@link #channel channel}. */
  public int getChannel() { return channel; }

  /** Sets the amount of {@link #channel channel}. Overwrite this method to
   * ensure code to be run in the constructor.
   * @param ch the amount of audio channels (at least one).
   * @see ALab.ASampleLoop ASampleLoop*/
  protected void setChannel(int ch) {
    channel = Math.max(ch, 1); last = channel; }

  /** Returns the next value ranging from <code>-1</code> to <code>1</code>.
   * The channel <code>ch</code> is a value between <code>0</code> and
   * <code>{@link #channel channel}-1</code>. If a non existing
   * channel (eg: channel number smaller than zero) is asked for any value,
   * zero is returned. The samplerate <code>rate</code> is any positive
   * integer. <br><br>
   * This method exists for developers help. It ensures that a value in ALab
   * is alway in the correct range, only valid channels are asked for values
   * and it remembers the last called channel, to notice whenever a new
   * sample starts. To make {@link #isNewSample(int) isNewSample()} work
   * correctly, ensure that each samples channels are always asked by
   * starting with the lowest channel number and ending with the highest one.
   * The method calculating the next value is {@link #getValue(int, int)
   * getValue()}.<br><br>
   *
   * @param ch the channel.
   * @param rate the samplerate.
   * @return the next value.
   * @exception AException if the value is not in the valid range or if
   *            another error occurs while calculating the next value.
   * @see #getValue(int, int) getValue()
   */
  public double nextValue(int ch, int rate) throws AException {
    if(ch >= channel) return 0;
    double v = getValue(ch, rate);
    if(Math.abs(v) > 1.0) throw new AException(
      this.getClass().getName()+".nextValue(): out of range (-1..1).");
    last = ch;
    return v; }

  /** Returns the next value of an object multiplied by another value
   * <code>v</code>. If the asked object does not exist, <code>v</code>
   * is returned. This method is used by most objects using modulators.
   * @param o the object asked for the next value.
   * @param ch the channel.
   * @param rate the samplerate.
   * @param v the multiplication value.
   * @return the next value multiplied by a value.
   * @exception AException if any calculation error occurs. */
  protected double nextObjValue(AObject o, int ch, int rate, double v) throws AException {
    if(o == null) return v; else return v * o.nextValue(ch, rate); }

  /** Marks weather a new sample has began or not.
   * @param ch the current channel number.
   * @return <code>true</code>: if the current channel is lower than or
   *         equal to the last called one, stored in {@link #last last}.
   *         <code>false</code>: if not. */
  protected boolean isNewSample(int ch) { return (ch <= last); }

  /** Returns an array of {@link #channel channel} <code>double</code>s
   * set to a certain value. An array of <code>channel</code> values is
   * created an initialized with <code>val</code>.
   * @param val the value.
   * @return the array. */
  protected double[] arrayWith(double val) {
    double[] v = new double[channel];
    for(int i = 0; i < channel; i++) v[i] = val;
    return v; }

  /** Returns the next calculated value.
   * @param ch the channel.
   * @param rate the samplerate.
   * @return the calculated value.
   * @exception AException if a calculation error occurs.
   * @see #nextValue(int, int) nextValue() */
  protected abstract double getValue(int ch, int rate) throws AException;

  /** Returns an exactly reconstruction of itself. All AObjects used
   * by this AObject are also copied before used in this new copy.
   * @return a copy of itself. */
  public abstract AObject copy();
}
