// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

import java.io.*;

/**
 * A class that does not calcualte its values, but reads its values
 * from an {@link ALab.AInputStream AInputStream} (raw PCM coded data).
 * These values can be played forward, backward or bidirectional. It is
 * also possible to play these values in a loop.<br><br>
 * ASample can only use AInputStreams that support marking. To encode
 * the values offered by the input stream it is neccessary to know, how
 * many bit are used to encode its values. ASample can handle bitranges
 * of 8, 16, 24 and 32 bit (is equal to 1, 2, 3 or 4 byte). Every
 * ASample has a certain start sample and a length. If values before or
 * after these positions are asked for, zero is returned. If the length
 * is negative, the samples are played backwards (beginning at start
 * position). <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see     ALab.AInputStream AInputStream
 * @see     ALab.ASampleLoop ASampleLoop
 */
public class ASample extends AObject {
  private AInputStream input;
  private int      bit;
  private int      bitMax;
  private int      sample;
  private long     start;
  private long     length;
  private long     max = 0;
  private long     pos = 0;
  private boolean  format;
  private boolean  reverse = false;
  private boolean  loop    = false;
  private boolean  bidir   = false;
  private boolean  bidone  = false;
  private boolean  init    = false;
  private double[] val;

  /**
   * Creates a new sample reading instance, that reads its
   * values from the specified AInputStream.
   *
   * @param ch  the amount of channels.
   * @param in  the input stream offering the samples to be read.
   * @param bit the bit range in byte (eg.
   *            <code>2</code> for bitrange of 16 Bit).
   * @param format the byte code format.
   *            <code>true</code>: big endian (motorola).
   *            <code>false</code>: little endian (intel).
   * @param start  the index of the first value to be read.
   * @param length the amount of samples to be read.
   */
  public ASample(int ch, AInputStream in, int bit,
                 boolean format, long start, long length) {
    super(ch); setInput(in); setBit(bit);
    setFormat(format); setStart(start); setLength(length); }

  /** Creates a new <code>ASample</code>
   * by copying the given <code>ASample</code>.
   * @param s the given object. */
  protected ASample(ASample s) {
    super(s);                setInput(s.getInput());
    setBit(s.getBit());      setFormat(s.getFormat());
    setStart(s.getStart());  setLength(s.getLength());
    setLoop(s.getLoop());    setBidir(s.getBidir()); }

  /** Returns the input stream. */
  public AInputStream getInput() { return input; }

  /** Sets the input stream to a copy of <code>in</code>.
   * @param     in the input stream. */
  public void setInput(AInputStream in) {
    try { input = in.copy(); }
    catch(IOException e) { input = null; }}

  /** Returns the maximal amount of samples that can be read
   * from the input stream. */
  public long getInputLength() { return max; }

  /** Returns the bitrange of the values of the input stream. */
  public int getBit() { return bit; }

  /**
   * Sets the bitrange of the values of the input stream. The
   * bitrange is set to <code>8*i</code>. <code>i</code> ranges
   * from <code>1</code> to <code>4</code>(8 to 32 Bit).
   * @param     i the bitrange in bytes. */
  public void setBit(int i) { bit = i;
    if(bit < 1) bit = 1; else if(bit > 4) bit = 4; }

  /** Returns the byte code format.
   * @return
   * <code>true</code>: big endian (motorola).
   * <code>false</code>: little endian (intel). */
  public boolean getFormat() { return format; }

  /** Sets the byte code format.
   * @param b
   * <code>true</code>: big endian (motorola).
   * <code>false</code>: little endian (intel). */
  public void setFormat(boolean b) { format = b; }

  /** Returns the index of the first sample to be read. */
  public long getStart() { return start; }

  /** Sets the index of the first sample to be read.
   * This index can also be set before or after the first or
   * the last sample that is offered by the input stream.
   * @param v the index of the first sample to be read. */
  public void setStart(long v) { start = v; }

  /** Returns the amount of samples to be read. */
  public long getLength() { return length; }

  /** Sets the amount of samples to be read. If this amount
   * is negative, the samples will be played backwards.
   * @param v the amount of samples to be read. */
  public void setLength(long v) { length = v; }

  /** Marks weather samples are played once or in a loop.
   * @return <code>true</code>: if samples are played in a loop.
   *         <code>false</code>: if samples are only played once. */
  public boolean getLoop() { return loop; }

  /** Turnes on or off the loop mode.
   * @param b <code>true</code>: samples are played in a loop.
   *          <code>false</code>: samples are only played once.
   */
  public void setLoop(boolean b) { loop = b; }

  /** Marks weather samples are played in one direction or in both.
   * @return
   * <code>true</code>: if samples are played in one direction.
   * <code>false</code>: if samples are played in both. */
  public boolean getBidir() { return bidir; }

  /** Turns on or off bidir mode.
   * @param b
   * <code>true</code>: samples are played in one direction.
   * <code>false</code>: samples are played in both. */
  public void setBidir(boolean b) { bidir = b; }

  private void init() throws AException { init = true;
    if(!input.markSupported()) throw new AException(
      "ASample.init(): input does not support mark.");
    try { input.reset(); }
    catch(IOException e) { input.mark(Integer.MAX_VALUE); }
    bitMax = 1 << (8*bit-1); sample = bit * channel;
    try { max = input.available() / sample; }
    catch(IOException e) { max = 0; }
    long v = length; length = Math.abs(length);
    if(v < 0) { reverse = true; start+= v; }
    if(reverse) set(start + length + 1); else set(start);
    val = arrayWith(0); }

  protected double getValue(int ch, int rate) throws AException {
    if(!init) init();
    if(!isNewSample(ch)) return val[ch];
    else if(!reverse) {
      if(pos >= (start + length))
        if(bidir && (!bidone || loop)) {
          bidone = true;
          reverse = true;
          set(start + length - 1); }
        else if(loop) set(start);
        else { val = arrayWith(0); return 0; }}
    else
      if(pos <= (start + 1))
        if(bidir && (!bidone || loop)) {
          bidone = true;
          reverse = false;
          set(start); }
        else if(loop) set(start + length - 1);
        else { val = arrayWith(0); return 0; }
      else set(pos - 2);
    if((pos < 0) || (pos >= max)) {
      set(pos + 1); val = arrayWith(0); return 0; }
    val = read(); return val[ch]; }

  /**
   * Returns an interpolated value of the input stream.
   * Reads at <code>pos</code> of the input stream and
   * interpolates between two values of an audio channel.
   * This value (between <code>-1</code> and <code>1</code>)
   * is returned.
   *
   * @param ch  the channel.
   * @param pos
   *   the position of the asked value in the inputstream.
   *   If this position is between two values of the input
   *   stream, the return value is calculated by linear
   *   interpolation.
   * @return the interpolated value.
   * @exception AException if an IOException occured.
   */
  public double readAt(int ch, double pos) throws AException {
    if(!init) init();
    byte[] b = new byte[bit];
    pos+= start;
    double f = pos % 1.0;
    double v1; double v2;
    long p = Math.round(Math.floor(pos));
    long l = p * sample + ch * bit;
    if(((p + 1) >= max) || (p < 0) || (p < start)
        || (p >= (start+length))) return 0;
    try {
      input.reset();
      while(l > 0) l-= input.skip(l);
      input.read(b, 0, bit);
      v1 = (double)byteArrayToInt(b, format) / bitMax;
      input.skip(sample - bit);
      input.read(b, 0, bit);
      v2 = (double)byteArrayToInt(b, format) / bitMax;
      input.skip(sample - (ch + 1) * bit); }
    catch(IOException e) { throw new AException(
      "ASample.read(): error reading from input."); }
    this.pos = p + 2;
    return f * v2 + (1 - f) * v1; }

  /**
   * Reads a whole sample. Reads at the current position of
   * the input stream and returns an array of values (between
   * <code>-1</code> and <code>1</code>).
   *
   * @return an array of values for each audio channel.
   * @exception AException
   *   if the index is outside the input stream
   *   or an IOException occured.
   */
  public double[] read() throws AException {
    double[] v = new double[channel];
    byte[] b = new byte[bit];
    if((pos >= max) || (pos < 0)) throw new AException(
      "ASample.read(): must not read outside the sample.");
    for(int i = 0; i < channel; i++) {
      try { input.read(b, 0, bit); }
      catch(IOException e) { throw new AException(
        "ASample.read(): error reading from input."); }
      v[i] = (double)byteArrayToInt(b, format) / bitMax; }
    pos++; return v; }

  /**
   * Sets the current position. Sets the position of the
   * input stream to any value. If the new position is
   * smaller than the last, the input streams
   * {@link ALab.AInputStream#reset() reset()} method
   * is called and then the position is set.
   *
   * @param pos the position the input stream should be set.
   * @exception AException if an IOException occurs.
   */
  public void set(long pos) throws AException {
    if((pos < 0) || (pos >= max)) { this.pos = pos; return; }
    boolean reset = false; long l;
    if(pos == this.pos) return;
    else if((pos > this.pos) && (this.pos >= 0))
      l = (pos - this.pos) * sample;
    else { l = pos * sample; reset = true; }
    try { this.pos = pos;
      if(reset) input.reset();
      while(l > 0) l-= input.skip(l); }
    catch(IOException e) { throw new AException(
      "ASample.set(): error reading from input."); }}

  private static int byteArrayToInt(byte[] b, boolean m) {
    int ret = 0;
    if(m) {
      if(b[0] < 0) ret = -1;
      for(int i = 0; i < b.length; i++)
        ret = (ret << 8) | (b[i] & 0xFF); }
    else {
      int start = b.length - 1;
      if(start == 0) return (byte)(b[0] - 128);
      if(b[start] < 0) ret = -1;
      for(int i = start; i >= 0; i--)
        ret = (ret << 8) | (b[i] & 0xFF); }
    return ret; }

  public AObject copy() { return new ASample(this); }
}