// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

/**
 * A class looping samplerate independently in an {@link
 * ALab.ASample ASample}. This class differs in some ways from
 * the class ASample. First of all it is a subclass of {@link
 * ALab.AWave AWave} (has a frequency, can do FM, ...).
 * ASampleLoop can not play a sample once, but loops all the
 * time. The loop start position and loop length is not
 * fixed and can be modulated. To archieve both - FM and
 * changing loop length - values are interpolated instead of
 * simply reading the samples. To avoid disturbing noise at
 * the end of the loop because of leveldifferences at start
 * and end position, ASampleLoop can generate smooth crossings,
 * by fading in and out the signal. The amount of smooth can
 * be set and modulated.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see     ALab.ASample ASample
 * @see     ALab.AWave AWave
 */
public class ASampleLoop extends AWave {
  private double  start;
  private double  length;
  private double  smooth = 0;
  private AObject startMod;
  private AObject lengthMod;
  private AObject smoothMod;
  private ASample input;
  private boolean bidir = false;
  private boolean[] dir;

  /**
   * Creates a new sample looping instance, that reads its
   * values from an ASample.
   *
   * @param in the ASample offering the samples to be read.
   * @param freq the frequency (in loops per second).
   * @param phase the phase.
   * @param start  the index of the first value to be read.
   * @param length the amount of samples to be read.
   */
  public ASampleLoop(ASample in, double freq,  double phase,
                           double start, double length) {
    super(in.getChannel(), freq, phase);
    setInput(in); setStart(start); setLength(length); }

  /** Creates a new <code>ASampleLoop</code>
   * by copying the given <code>ASampleLoop</code>.
   * @param o the given object. */
  protected ASampleLoop(ASampleLoop o) {  super(o);
    setStart(o.getStart());   setStartMod(o.getStartMod());
    setLength(o.getLength()); setLengthMod(o.getLengthMod());
    setSmooth(o.getSmooth()); setSmoothMod(o.getSmoothMod());
    setInput(o.getInput());   setBidir(o.getBidir()); }

  protected void setChannel(int ch) {
    super.setChannel(ch); dir = new boolean[channel];
    for(int i = 0; i < channel; i++) dir[i] = true; }

  /** Returns the index of the first sample to be read. */
  public double getStart() { return start; }

  /** Sets the index of the first sample to be read.
   * This index can also be set before or after the first or
   * the last sample that is offered by the ASample.
   * @param v the index of the first sample to be read. */
  public void setStart(double v) { start = v; }

  /** Returns the start modulator. */
  public AObject getStartMod() { return startMod; }

  /** Sets the start modulator to a copy of <code>o</code>.
   * @param o the start modulator. */
  public void setStartMod(AObject o) {
    if(o != null) startMod = o.copy(); }

  /** Returns the amount of samples a loop has. */
  public double getLength() { return length; }

  /** Sets the amount of samples a loop has. If this amount
   * is negative, the samples will be played backwards.
   * @param v the amount of samples. */
  public void setLength(double v) { length = v; }

  /** Returns the length modulator. */
  public AObject getLengthMod() { return lengthMod; }

  /** Sets the length modulator to a copy of <code>o</code>.
   * @param o the length modulator. */
  public void setLengthMod(AObject o) {
    if(o != null) lengthMod = o.copy(); }

  /** Returns the ASample offering the sample values. */
  public ASample getInput() { return input; }

  /** Sets the input ASample to a copy of <code>o</code>.
   * @param in the ASample offering the sample values. */
  public void setInput(ASample o) {
    input = (ASample)o.copy(); }

  /** Marks weather samples are played in one direction or in both.
   * @return
   * <code>true</code>: if samples are played in one direction.
   * <code>false</code>: if samples are played in both. */
  public boolean getBidir() { return bidir; }

  /** Turns on or off bidir mode.
   * @param b
   * <code>true</code>: samples are played in one direction.
   * <code>false</code>: samples are played in both. */
  public void setBidir(boolean b) { bidir = b; }

  /** Returns the phase duration for smooth. */
  public double getSmooth() { return smooth; }

  /** Sets the phase duration for smooth. At the crossing of
   * start and end of the loop, values are faded in and out.
   * Smooth values higher than <code>PI</code> or lower than
   * <code>0</code> are truncated to the valid range.
   * @param v the phase duration for fading. */
  public void setSmooth(double v) {
    smooth = Math.min(Math.abs(v), Math.PI); }

  /** Returns the smooth modulator. */
  public AObject getSmoothMod() { return smoothMod; }

  /** Sets the smooth modulator to a copy of <code>o</code>.
   * @param o the smooth modulator. */
  public void setSmoothMod(AObject o) {
    if(o != null) smoothMod = o.copy(); }

  protected double getValue(int ch, int rate) throws AException {
    double pos; double f;
    double v = nextPhase(ch, rate);
    double s = nextObjValue( startMod, ch, rate,  start);
    double l = nextObjValue(lengthMod, ch, rate, length);
    double m = nextObjValue(smoothMod, ch, rate, smooth);
    if(Math.abs(v) < m) f = Math.abs(v/smooth); else f = 1;
    if(v < 0) v+= PIx2;
    if(bidir && isZero(ch)) dir[ch] = !dir[ch];
    if(bidir && dir[ch]) pos = s + (1 - v / PIx2) * l;
    else                 pos = s +      v / PIx2  * l;
    return input.readAt(ch, pos) * f; }

  public AObject copy() { return new ASampleLoop(this); }
}