// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

import java.io.*;

/**
 * A class storing values to disc for later use. If you don't want the
 * values being calculated every time, you can store the values to disc
 * and use them as if they were just calculated. If the values of the
 * file have become old you can update the values. The length of the
 * stored file tells how many samples will be precalculated and stored
 * to disc. If any value outside this length is asked for, zero is
 * returned. Eg.
 * <pre>
 *   AObject obj = complexObject; // any complex calculated AObject
 *   AStore  obj = new AStore(obj, 88200, 44100, 2, "ALab.tmp", false);
 *   AObject complexerObject = new AComplex1(obj);
 *   ...
 * </pre>
 * The above example would also work, if the line with the
 * <code>AStore</code> command would be deleted. But the difference
 * is, that two seconds of the <code>complexObject</code> are stored
 * to disc. If the code is compiled again, this time with eg.
 * <code>AComplex2(obj)</code> instead of <code>AComplex1(obj)</code>
 * then the stored values of <code>complexObject</code> would be
 * used instead of calculating them again.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see     ALab.ABuffer ABuffer
 */
public class AStore extends AObject {
  private ASample file;

  /** Creates a store object that precalculates and stores length
   * samples of an AObject to a file. If the file already exists
   * and the file should not be updated, the values in this file
   * will be returned.
   * @param o the AObject with the values to be stored.
   * @param length the amount of sample s to be stored.
   * @param rate the samplerate of the calculated values.
   * @param bit the bitrate in byte (eg. two for 16Bit).
   * @param name the name of the file.
   * @param update <code>true</code>: the values are precalculated even
   *   if the file already exists. <code>false</code>: the values are
   *   precalculated only if the file doesn't exist.
   * @exception AException if any calculation error occurs. */
  public AStore(AObject o, long length, int rate,
                int bit, String name, boolean update) throws AException {
    super(o);
    if(!(new File(name)).exists() || update) try {
      AContainer c = new AContainer(channel);
      c.append(o, length);
      BufferedOutputStream out = null;
      out = new BufferedOutputStream(new FileOutputStream(name));
      c.write(rate, bit, true, out);
      out.close(); }
      catch(IOException e) { throw new AException(
        "AStore(): could not open file for writing."); }
    try { file = new ASample(channel, new AFile(name), bit, true, 0, length);
    } catch(IOException e) { throw new AException(
      "AStore(): could not open file for reding."); }}

  /** Creates a new <code>AStore</code> by copying the given
   * <code>AStore</code> using the given stored file.
   * @param o the given object. */
  protected AStore(AStore o) { super(o); setFile(o.getFile()); }

  /** Returns the {@link ALab.ASample ASample}
   *  storing the precalculated values. */
  public ASample getFile() { return file; }

  /** Sets the {@link ALab.ASample ASample} storing the
   * precalculated values to a copy of <code>o</code>.
   * @param o the object. */
  protected void setFile(ASample o) { file = (ASample)o.copy(); }

  protected double getValue(int ch, int rate) throws AException {
    return file.nextValue(ch, rate); }

  public AObject copy() { return new AStore(this); }
}