// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

import java.io.*;
import java.util.*;

/**
 * A class acting as a table for any value between <code>-1</code> and
 * <code>1</code>. This class is not an {@link ALab.AObject AObject}.
 * It transformes values into other values by using a table. The method
 * {@link #valueAt(double) valueAt(pos)} returns the value of the table
 * at <code>pos</code>. Both - <code>pos</code> and the returned value -
 * have to be inside the range <code>-1..1</code>. In the table there are
 * only defined values at positions above and equal zero. Positions lower
 * than zero are calculated by returning the negative value of the result
 * at the positve position (eg: <code>valueAt(0.5) = 1</code> than
 * <code>valueAt(-0.5) = -1</code>). All values are calculated by linear
 * interpolating the table values. The size of the table defines how many
 * values are used to represent the table. <br><br>
 * <IMG SRC="space.gif" WIDTH=50 HEIGHT=1>
 * <IMG SRC="ATable.gif" ALT="Diagram of some tables."><br>
 * Tables can be created by reading the values from an input stream or by
 * reading the values from an AObject. To store the table the values can
 * be written to an output stream.<br><br>
 * Notice that tables can not be copied. If you use a table by any other
 * object and change the table afterwards, the table used by the object
 * will also change! <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see     ALab.ATablelize ATablelize
 */
public class ATable {
  private double[] table;
  private int size;

  /**
   * Creates a new table by reading the values from an input stream. The
   * format of the values represented by the input stream must be strings
   * that define a certain double value (eg. <code>0.24</code>), or
   * quoted strings (eg. <code>'-0.5'</code> or <code>'0.1E-4'</code>).
   * A table file must at least have two values. eg:
   * <pre>
   * // table file
   * // java style comments are ignored
   * 0.0  0.1  0.2  0.3  0.4
   * 0.5  0.59 0.66 0.73 0.77 0.8
   * // end file
   * </pre>
   *
   * @param in the input stream representing the values.
   * @exception AException if less than two values are represented or any
   *   value is out of range (<code>-1..1</code>) or any IOError occurs.
   */
  public ATable(InputStream in) throws AException { try {
    StreamTokenizer tok = new StreamTokenizer(new InputStreamReader(in));
    Vector v = new Vector();
    while(tok.nextToken() != tok.TT_EOF)
      if(tok.ttype == tok.TT_NUMBER) v.addElement(new Double(tok.nval));
      else v.addElement(Double.valueOf(tok.sval));
    if(v.size() > 1) { size = v.size(); table = new double[size];
      for(int i = 0; i < size; i++) {
        table[i] = ((Double)v.elementAt(i)).doubleValue();
        if(Math.abs(table[i]) > 1) throw new AException(
          "ATable(): out of range (-1..1)."); }}
    else throw new AException("ATable(): too less valid values."); }
    catch(IOException e) { throw new AException(
      "ATable(): error parsing InputStream."); }}

  /**
   * Creates a new table by reading the values from an {@link
   * ALab.AObject AObject}. The values are read from the first channel
   * (channel zero).
   *
   * @param o the object representing the values.
   * @param length the amount of values.
   * @param rate the samplerate the values are calculated.
   * @exception AException if any calculation error occurs.
   */
  public ATable(AObject o, int length, int rate) throws AException {
    size = Math.max(2, length); table = new double[size];
    for(int i = 0; i < size; i++) table[i] = o.nextValue(0, rate); }

  /** Returns the interpolated value at a certain table position.
   * @param pos the position in the table (<code>-1..1</code>).
   * @return the interpolated value. */
  public double valueAt(double pos) {
    boolean neg = (pos < 0); pos = Math.abs(pos);
    if(pos > 1) return 0; pos*= (size-1);
    int index = (int)Math.floor(pos); pos = pos % 1.0;
    if(pos == 0) pos = table[index];
    else pos = table[index] * (1-pos) + table[index+1] * pos;
    if(neg) return -pos; else return pos; }

  /** Writes the values to an output stream.
   * @param out the output stream. */
  public void write(OutputStream out) {
    PrintStream p = new PrintStream(out);
    p.println("// ALab-Table: "+size+" values:");
    for(int i = 0; i < size; i++) p.println("'"+table[i]+"'");
    p.close(); }
}
