// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

/**
 * A class transforming an {@link ALab.AObject AObject}s values depending on two
 * tables. Instead of the AObjects values the {@link ALab.ATable ATable}s values
 * are returned. If the mix value is <code>1</code> then the values of table one
 * are returned. If the mix value is <code>-1</code> then the values of table two
 * are returned. Values between <code>-1</code> and <code>1</code> will result in
 * a mix of both tables. Eg: two tables, one linear, the other linear at values
 * below <code>0.5</code> and flat for values above <code>0.5</code>. <br><br>
 * <IMG SRC="space.gif" WIDTH=50 HEIGHT=1>
 * <IMG SRC="ATablelize.gif" ALT="Diagram of the tablelizing effect."><br>
 * In the first case, the AObjects values will not be changed. In the last case
 * all values above <code>0.5</code> will be truncated to <code>0.5</code> and
 * all values below <code>-0.5</code> truncated to <code>-0.5</code>.
 * The mix value can be modulated.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 * @see     ALab.ATable ATable
 */
public class ATablelize extends AObject {
  private double  mix;
  private AObject mixMod = null;
  private AObject obj;
  private ATable  tab1;
  private ATable  tab2;

  /** Creates a value by table transforming instance. If the second
   * table is null, it is asumed to be linear.
   * @param o the values of this object will be transformed.
   * @param tab1 the first table.
   * @param tab2 the second table or <CODE>null</CODE> for linear.
   * @param mix the mixing value to interpolate between both tables.
   * @see ALab.ATable ATable */
  public ATablelize(AObject o, ATable tab1, ATable tab2, double mix) {
    super(o); setObj(o);
    setTable1(tab1); setTable2(tab2); setMix(mix); }

  /** Creates a new <code>ATablelize</code>
   * by copying the given <code>ATablelize</code>.
   * @param o the given object. */
  protected ATablelize(ATablelize o) {
    super(o);                 setObj(o.getObj());
    setTable1(o.getTable1()); setTable2(o.getTable2());
    setMix(o.getMix());       setMixMod(o.getMixMod()); }

  /** Returns the mixing value. */
  public double getMix() { return mix; }

  /** Sets the mixing value.
   * @param v the mixing value between <code>-1</code>(table two)
              and <code>1</code>(table one). */
  public void setMix(double v) { mix = v; }

  /** Returns the mix modulator. */
  public AObject getMixMod() { return mixMod; }

  /** Sets the mix modulator to a copy of <code>o</code>.
   * @param o the mix modulator. */
  public void setMixMod(AObject o) {
    if(o != null) mixMod = o.copy(); }

  /** Returns the object that will be transformed. */
  public AObject getObj() { return obj; }

  /** Sets the object that will be transformed to a copy of <code>o</code>.
   * @param o the values of this object will be transforemd. */
  protected void setObj(AObject o) { obj = o.copy(); }

  /** Returns the first table. */
  public ATable getTable1() { return tab1; }

  /** Sets the first table.
   * @param table the table*/
  protected void setTable1(ATable table) { tab1 = table; }

  /** Returns the second table. */
  public ATable getTable2() { return tab2; }

  /** Sets the second table.
   * @param table the table*/
  protected void setTable2(ATable table) { tab2 = table; }

  protected double getValue(int ch, int rate) throws AException {
    double v = obj.nextValue(ch, rate);
    double f = nextObjValue(mixMod, ch, rate, mix)/2 + 0.5;
    if(tab2 == null) return tab1.valueAt(v) * f + v * (1 - f);
    else return tab1.valueAt(v) * f + tab2.valueAt(v) * (1 - f); }

  public AObject copy() { return new ATablelize(this); }
}