// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

package ALab;

/**
 * Abstract basisclass for all classes having a frequency.
 * Subclasses have to implement the abstract methods of
 * <code>{@link ALab.AObject AObject}</code>. The frequency
 * should never be set higher than the half samplerate.
 * <br><br>
 * <b>Freqency Modulation:</b> in ALab modulation always works
 * the same way. there is a certain value(eg. <code>freq</code>)
 * and an {@link ALab.AObject AObject} (eg.
 * <code>freqMod</code>) that modulates this value. This means
 * that the modulated values range is between the positive and
 * the negative value (eg. <code>-freq</code> ...
 * <code>freq</code>). If you want the value be modulated
 * between zero and its value, the modulation-objects values
 * must not be lower than zero.<br>
 * The class {@link ALab.AAmp AAmp} can be used to do this.
 * The following example shows how to modulate the frequency of
 * a 440Hz sinus by a 10Hz sinus. The modulated frequency ranges
 * from 0 to 440Hz.
 * <pre>
 *    AWave sin = new ASin(1, 440, 0);
 *    AWave mod = new ASin(1,  10, 0);
 *    sin.setFreqMod(new AAmp(mod, 0.5, 0.5)); // mod * 0.5 + 0.5
 * </pre>
 * <b>Phase Modulation:</b> works the same way as FM but
 * modulates the phase.
 * <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 */
public abstract class AWave extends AObject {
  /** {@link java.lang.Math.PI Math.PI} multiplied by <code>2.0</code> */
  protected static double PIx2 = 2 * Math.PI;
  private double    freq;
  private double    phase;
  private AObject   freqMod  = null;
  private AObject   phaseMod = null;
  private double[]  pos;     // marks current pos in wave
  private double[]  old;     // marks last pos in wave
  private double[]  oldP;    // marks last phase change
  private boolean[] zero;

  /** Creates a new wave object with a certain frequency and phase.
   * @param ch    the amount of channels.
   * @param freq  the frequency.
   * @param phase the phase. */
  protected AWave(int ch, double freq, double phase) {
    super(ch); setFreq(freq); setPhase(phase); }

  /** Creates a new wave object by copying the given object.
   * @param w the given object. */
  protected AWave(AWave w) {    super(w);
    setFreq(w.getFreq());       setPhase(w.getPhase());
    setFreqMod(w.getFreqMod()); setPhaseMod(w.getPhaseMod()); }

  /** Returns the frequency. */
  public double getFreq() { return freq; }

  /** Sets the frequency.
   * @param v the frequency in Hz. */
  public void setFreq(double v) { freq = v; }

  /** Returns the frequency modulator. */
  public AObject getFreqMod() { return freqMod; }

  /** Sets the frequency modulator to a copy of <code>o</code>.
   * @param o the frequency modulator. */
  public void setFreqMod(AObject o) {
    if(o != null) freqMod = o.copy(); }

  /** Returns the phase. */
  public double getPhase() { return phase; }

  /** Sets the phase. The value of phase ranges from
   * <code>(-2 * PI)</code> to <code>(2 * PI)</code>.
   * Values out of this range are truncated.
   * @param v the phase in radians. */
  public void setPhase(double v) {
    if(v> PIx2) v = PIx2; else if(v< -PIx2) v = -PIx2;
    phase = v; pos = arrayWith(v); old = arrayWith(v);
    oldP = arrayWith(v); zero = new boolean[channel]; }

  /** Returns the phase modulator. */
  public AObject getPhaseMod() { return phaseMod; }

  /** Sets the phase modulator to a copy of <code>o</code>.
   * @param o the phase modulator. */
  public void setPhaseMod(AObject o) {
    if(o != null) phaseMod = o.copy(); }

  /** Marks beginning of a new circle.
   * If the phase at this certain channel has just crossed
   * zero, then a new circle has began at this channel.
   * This method needs the result of the method
   * {@link #nextPhase(int, int) nextPhase()} and so must
   * be called after it. To ensure this, you must call
   * one of the following methods before:
   * {@link #nextValue(int, int) nextValue()},
   * {@link #getValue(int, int) getValue()}
   * or {@link #nextPhase(int, int) nextPhase()}
   *
   * @param     ch the channel.
   * @return    <code>true</code>: if phase has crossed zero.
   *            <code>false</code>: otherwise.
   * @see       ALab.AWaveTrigger AWaveTrigger
   */
  public boolean isZero(int ch) { return zero[ch]; }

  /** Calculates the next phase depending on frequency,
   * phase, frequency modulation and phase modulation.
   * @param  ch   the channel.
   * @param  rate the samplerate.
   * @return a value between <code>-PI</code> and <code>PI</code>.
   *         A new circle starts with <code>0</code>.
   * @exception AException if any calculation error occurs.
   */
  protected double nextPhase(int ch, int rate) throws AException {
    double f = nextObjValue( freqMod, ch, rate,  freq) * PIx2;
    double p = nextObjValue(phaseMod, ch, rate, phase) - oldP[ch];
    zero[ch] =           (((old[ch] < 0) && (pos[ch] > 0))
      || (pos[ch] == 0) ||((old[ch] > 0) && (pos[ch] < 0)));
    if     (pos[ch] >  Math.PI) pos[ch] -= PIx2;
    else if(pos[ch] < -Math.PI) pos[ch] += PIx2;
    old[ch] = pos[ch];
    double v = pos[ch];
    pos[ch]+= (f / rate) + p;
    oldP[ch]+= p;
    if     (v >  Math.PI) v-= PIx2;
    else if(v < -Math.PI) v+= PIx2;
    return v; }
}