// ALab v1.0: java package for creating experimental digital sounds and soundeffects
// Copyright (C) 2001 Michael Bliem   ALab@gmx.at      see 'license.txt' for details

import ALab.*;

/**
 * A two channel class that sets the stereo position of an object.
 * The panorama value <code>pan</code> controlls the position of
 * the values in the stereo area. If <code>pan</code> is zero
 * the object is set to the middle of the panorama, by routing
 * the half amplituded values to both channels. If <code>pan</code>
 * is <code>1</code> the full amplituded value is routed to channel
 * one (right). If <code>pan</code> is <code>-1</code> the values
 * are routed to channel zero (left). The panorama value can be
 * modulated (for each channel). <br><br><IMG SRC="point.gif">
 *
 * @author  Michael Bliem
 * @version 1.0
 */
public class APan extends AObject {
  private double   pan;            // the pan value
  private AObject  panMod = null;  // the pan value modulator
  private AObject  obj;            // the AObject to get a pan
  private double[] smpVal;         // array to store one sample
  private double[] panVal;         // array to store pan values

  /** Creates a new panorama instance.
   * @param o the object that's stereo position is set.
   * @param pan the panorama value. */
  public APan(AObject o, double pan) { super(2);
    setObj(o); setPan(pan); }

  /** Creates a new <code>APan</code>
   * by copying the given <code>APan</code>.
   * @param o the given object. */
  protected APan(APan o) { super(2);
    setObj(o.getObj());
    setPan(o.getPan());
    setPanMod(o.getPanMod()); }

  /** Returns the panorama value. */
  public double getPan() { return pan; }

  /** Sets the panorama value (range: <code>-1..1</code>).
   * @param v the panorama value. */
  public void setPan(double v) {
    if(v < -1) v = -1; else if(v > 1) v = 1;
    pan = v; }

  /** Returns the panorama modulator. */
  public AObject getPanMod() { return panMod; }

  /** Sets the panorama modulator to a copy of <code>o</code>.
   * @param o the panorama modulator. */
  public void setPanMod(AObject o) {
    if(o != null) panMod = o.copy(); }

  /** Returns the object that's stereo position is set. */
  public AObject getObj() { return obj; }

  /** Sets the object that's stereo position is set
   * to a copy of <code>o</code>.
   * @param o the object. */
  protected void setObj(AObject o) { obj = o.copy();
    smpVal = new double[obj.getChannel()];
    panVal = new double[obj.getChannel()]; }

  protected double getValue(int ch, int rate) throws AException {

    // if a new sample starts, all values of the sample
    // and all (modulated) panorama values are stored
    if(isNewSample(ch))

      // one sample of the object to be set in the panorama
      // consists of getChannel() channels.
      for(int i = 0; i < obj.getChannel(); i++) {

        // the value of every channel is stored
        smpVal[i] = obj.nextValue(i, rate);

        // the (modulated) panorama value of each channel
        // is stored (after divided by two).
        panVal[i] = nextObjValue(panMod, i, rate, pan) / 2.0; }

    // end of code only done at new sample
    // (following code runs for every channel)

    // result initialised with zero
    double result = 0;

    // the result is a mix of all channels of the object
    // that is to be set in the stereo panorama
    // this object has getChannel() channels
    for(int i = 0; i < obj.getChannel(); i++)

      // if left channel was asked result is incremented
      // by sample value amplified with left panorama amount
      if(ch == 0) result += smpVal[i] * (0.5 - panVal[i]);

      // if right channel was asked result is incremented
      // by sample value amplified with right panorama amount
      else        result += smpVal[i] * (0.5 + panVal[i]);

    // result is returned (overflow might occur!)
    return result; }

  public AObject copy() { return new APan(this); }
}

